import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.aerodynamics.flops_based.ground_effect import GroundEffect
from aviary.utils.aviary_values import AviaryValues, get_items
from aviary.variable_info.variables import Aircraft, Dynamic


class TestGroundEffect(unittest.TestCase):
    '''
    Perform regression test in all three computational ranges:
    - in ground effect (on the ground);
    - in transition (in the air, but near enough to the ground for limited effect);
    - in free flight (no ground effect).
    '''

    def test_regression(self):
        prob: om.Problem = make_problem()

        prob.run_model()

        tol = 1e-12

        for (key, (desired, units)) in get_items(_regression_data):
            try:
                actual = prob.get_val(key, units)

                assert_near_equal(actual, desired, tol)

            except ValueError as error:
                msg = f'"{key}": {error!s}'

                raise ValueError(msg) from None

        partials = prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partials, atol=1e-10, rtol=1e-12)


def make_problem():
    '''
    Return a problem that covers all three computational ranges:
    - in ground effect (on the ground);
    - in transition (in the air, but near enough to the ground for limited effect);
    - in free flight (no ground effect).
    '''
    nn = 3
    ground_altitude = 100.  # units='m'

    minimum_drag_coefficient = (0.02036, 'unitless')
    base_lift_coefficient = (np.array([0.5, 1.2, 2.0]), 'unitless')
    base_drag_coefficient = (np.array([0.02036, 1.1 * 0.02036, .08]), 'unitless')
    aspect_ratio = (9.45, 'unitless')
    height = (8., 'ft')
    span = (34., 'm')

    inputs = AviaryValues({
        'angle_of_attack': (np.array([0., 2., 6]), 'deg'),
        Dynamic.Mission.ALTITUDE: (np.array([100.0, 132, 155]), 'm'),
        Dynamic.Mission.FLIGHT_PATH_ANGLE: (np.array([0., 0.5, 1.0]), 'deg'),
        'minimum_drag_coefficient': minimum_drag_coefficient,
        'base_lift_coefficient': base_lift_coefficient,
        'base_drag_coefficient': base_drag_coefficient,
        Aircraft.Wing.ASPECT_RATIO: aspect_ratio,
        Aircraft.Wing.HEIGHT: height,
        Aircraft.Wing.SPAN: span
    })

    ground_effect = GroundEffect(num_nodes=nn, ground_altitude=ground_altitude)

    prob = om.Problem()
    prob.model.add_subsystem('ground_effect', ground_effect, promotes=['*'])
    prob.setup(force_alloc_complex=True)

    for (key, (val, units)) in inputs:
        prob.set_val(key, val, units)

    return prob


# NOTE:
# - results from `generate_regression_data()`
#    - last generated 2023 June 7
# - generate new regression data if, and only if, ground effect is updated with a more
#   trusted implementation
_regression_data = AviaryValues(dict(
    lift_coefficient=([0.5582336522780803, 1.200675778095871, 2.0], 'unitless'),
    drag_coefficient=([0.02036, 0.022395716686151382, 0.08], 'unitless')))


def generate_regression_data():
    '''
    Generate regression data that covers all three computational ranges:
    - in ground effect (on the ground);
    - in transition (in the air, but near enough to the ground for limited effect);
    - in free flight (no ground effect).

    Notes
    -----
    Use this function to generate new regression data if, and only if, ground effect is
    updated with a more trusted implementation.
    '''
    prob: om.Problem = make_problem()

    prob.run_model()

    lift_coefficient = prob.get_val('lift_coefficient')
    drag_coefficient = prob.get_val('drag_coefficient')

    prob.check_partials(compact_print=True, method="cs")

    print('lift_coefficient', *lift_coefficient, sep=', ')
    print('drag_coefficient', *drag_coefficient, sep=', ')


if __name__ == "__main__":
    unittest.main()



import unittest

import openmdao.api as om

from aviary.mission.ode.specific_energy_rate import SpecificEnergyRate
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_data.flops_data.full_mission_test_data import \
    data
from aviary.validation_cases.validation_tests import do_validation_test
from aviary.variable_info.variables import Dynamic


class SpecificEnergyRateTest(unittest.TestCase):
    def setUp(self):
        prob = self.prob = om.Problem()

        time, _ = data.get_item('time')

        prob.model.add_subsystem(
            "specific_energy_rate",
            SpecificEnergyRate(num_nodes=len(time)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        do_validation_test(self.prob,
                           'full_mission_test_data',
                           input_validation_data=data,
                           output_validation_data=data,
                           input_keys=[Dynamic.Mission.DRAG,
                                       Dynamic.Mission.MASS,
                                       Dynamic.Mission.THRUST_TOTAL,
                                       Dynamic.Mission.VELOCITY],
                           output_keys=Dynamic.Mission.SPECIFIC_ENERGY_RATE,
                           tol=1e-12)

    # TODO IO test will fail until mission variable hirerarchy implemented
    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



import openmdao.api as om

from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic


class LiftAndDragIncrements(om.ExplicitComponent):
    def setup(self):

        # Inputs

        self.add_input(
            "VDEL5",
            val=0.90761,
            units='unitless',
            desc="VDEL5: sensitivity of flap minimum drag coefficient to flap hinge line sweep",
        )
        self.add_input(
            "VLAM8",
            val=0.74444,
            units='unitless',
            desc="VLAM8: sensitivity of flap clean wing maximum lift coefficient to wing sweep angle",
        )
        self.add_input(
            "VDEL4",
            val=0.93578,
            units='unitless',
            desc="VDEL4: sensitivity of minimum drag coefficient to fuselage width to span ratio",
        )
        add_aviary_input(self, Aircraft.Wing.FLAP_DRAG_INCREMENT_OPTIMUM, val=0.1)
        self.add_input(
            "VDEL1",
            val=1.0,
            units='unitless',
            desc="VDEL1: sensitivity of flap minimum drag coefficient to flap chord ratio",
        )
        self.add_input(
            "VDEL2",
            val=0.62455,
            units='unitless',
            desc="VDEL2: sensitivity of flap minimum drag coefficient to flap angle",
        )
        self.add_input(
            "VDEL3",
            val=0.765,
            units='unitless',
            desc="VDEL3: sensitivity of flap minimum drag coefficient to partial flap span",
        )
        add_aviary_input(self, Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM, val=1.5)
        self.add_input(
            "VLAM3",
            val=0.97217,
            units='unitless',
            desc="VLAM3: sensitivity of flap clean wing maximum lift coefficient to aspect ratio",
        )
        self.add_input(
            "VLAM4",
            val=1.25725,
            units='unitless',
            desc="VLAM4: sensitivity of flap clean wing maximum lift coefficient slope to wing thickness",
        )
        self.add_input(
            "VLAM5",
            val=1,
            units='unitless',
            desc="VLAM5: sensitivity of flap clean wing maximum lift coefficient to wing flap to chord ratio",
        )
        self.add_input(
            "VLAM6",
            val=1,
            units='unitless',
            desc="VLAM6: sensitivity of flap clean wing maximum lift coefficient to wing flap deflection",
        )
        self.add_input(
            "VLAM7",
            val=0.735,
            units='unitless',
            desc="VLAM7: sensitivity of flap clean wing maximum lift coefficient to wing flap span",
        )
        self.add_input(
            "VLAM13", val=1.03512, units='unitless', desc="VLAM13: reynolds number correction factor"
        )
        self.add_input(
            "VLAM14", val=0.99124, units='unitless', desc="VLAM14: mach number correction factor "
        )

        # outputs

        self.add_output("delta_CD", val=0.0,
                        desc="DCD: increment on drag coefficient", units="unitless")
        self.add_output("delta_CL", val=0.0,
                        desc="DCL: increment on lift coefficient", units="unitless")

    def setup_partials(self):

        self.declare_partials(
            "delta_CD",
            [Aircraft.Wing.FLAP_DRAG_INCREMENT_OPTIMUM,
                "VDEL1", "VDEL2", "VDEL3", "VDEL4", "VDEL5"],
            dependent=True,
            method="cs",
            step=1e-8,
        )
        self.declare_partials(
            "delta_CL",
            [
                Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM,
                "VLAM3",
                "VLAM4",
                "VLAM5",
                "VLAM6",
                "VLAM7",
                "VLAM8",
                "VLAM13",
                "VLAM14",
            ],
            dependent=True,
            method="cs",
            step=1e-8,
        )

    def compute(self, inputs, outputs):

        delta_drag_trailing = inputs[Aircraft.Wing.FLAP_DRAG_INCREMENT_OPTIMUM]
        trailing_lift_increment = inputs[Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM]
        VDEL1 = inputs["VDEL1"]
        VDEL2 = inputs["VDEL2"]
        VDEL3 = inputs["VDEL3"]
        VDEL4 = inputs["VDEL4"]
        VDEL5 = inputs["VDEL5"]
        VLAM3 = inputs["VLAM3"]
        VLAM4 = inputs["VLAM4"]
        VLAM5 = inputs["VLAM5"]
        VLAM6 = inputs["VLAM6"]
        VLAM7 = inputs["VLAM7"]
        VLAM8 = inputs["VLAM8"]
        VLAM13 = inputs["VLAM13"]
        VLAM14 = inputs["VLAM14"]

        outputs["delta_CD"] = delta_CD = (
            delta_drag_trailing * VDEL1 * VDEL2 * VDEL3 * VDEL4 * VDEL5
        )
        outputs["delta_CL"] = delta_CL = (
            trailing_lift_increment
            * VLAM3
            * VLAM4
            * VLAM5
            * VLAM6
            * VLAM7
            * VLAM8
            * VLAM13
            * VLAM14
        )



import numpy as np

from openmdao.api import Group, MetaModelStructuredComp


# These battery values are entirely fabricated and not representative of any real battery.
# This model is used as an example of how to integrate external subsystems in Aviary.
T_bp = np.array([0., 60.])
SOC_bp = np.linspace(0., 1., 32)

tU_oc = np.vstack((np.linspace(2.8, 4.2, 32), np.linspace(2.7, 4.1, 32)))
tC_Th = np.ones((2, 32)) * 1000.0
tR_Th = np.vstack((np.linspace(0.06, 0.03, 32), np.linspace(0.04, 0.02, 32)))
tR_0 = np.vstack((np.linspace(0.06, 0.03, 32), np.linspace(0.04, 0.02, 32)))


class RegTheveninInterpGroup(Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        n = self.options['num_nodes']

        thevenin_interp_comp = MetaModelStructuredComp(method='slinear', vec_size=n,
                                                       training_data_gradients=False, extrapolate=True)

        thevenin_interp_comp.add_input('T_batt', 1.0, T_bp, units='degC')
        thevenin_interp_comp.add_input('SOC', 1.0, SOC_bp, units=None)
        thevenin_interp_comp.add_output('C_Th', 1.0, tC_Th, units='F')
        # add R_th and R_0 to get total R
        thevenin_interp_comp.add_output('R_Th', 1.0, tR_Th, units='ohm')
        thevenin_interp_comp.add_output('R_0', 1.0, tR_0, units='ohm')
        thevenin_interp_comp.add_output('U_oc', 1.0, tU_oc, units='V')

        self.add_subsystem(name='interp_comp',
                           subsys=thevenin_interp_comp,
                           promotes=["*"])  # promots params and values



import subprocess
import unittest
from pathlib import Path

import pkg_resources
from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs


@use_tempdirs
class CommandEntryPointsTestCases(unittest.TestCase):

    def run_and_test_cmd(self, cmd):
        # this only tests that a given command line tool returns a 0 return code. It doesn't
        # check the expected output at all.  The underlying functions that implement the
        # commands should be tested seperately.
        try:
            output = subprocess.check_output(cmd.split())
        except subprocess.CalledProcessError as err:
            self.fail(f"Command '{cmd}' failed.  Return code: {err.returncode}")

    @require_pyoptsparse(optimizer="SNOPT")
    def bench_test_SNOPT_cmd(self):
        cmd = 'aviary run_mission models/test_aircraft/aircraft_for_bench_GwGm.csv --optimizer SNOPT --max_iter 1'
        self.run_and_test_cmd(cmd)

    @require_pyoptsparse(optimizer="IPOPT")
    def bench_test_IPOPT_cmd(self):
        cmd = 'aviary run_mission models/test_aircraft/aircraft_for_bench_GwGm.csv --optimizer IPOPT --max_iter 1'
        self.run_and_test_cmd(cmd)

    @require_pyoptsparse(optimizer="IPOPT")
    def bench_test_IPOPT_cmd(self):
        cmd = 'aviary run_mission models/test_aircraft/aircraft_for_bench_GwGm.csv' \
            ' --optimizer IPOPT --max_iter 1 --shooting'
        self.run_and_test_cmd(cmd)

    def test_diff_configuration_conversion(self):
        filepath = pkg_resources.resource_filename('aviary',
                                                   'models/test_aircraft/converter_configuration_test_data_GwGm.dat')
        outfile = Path.cwd() / 'test_aircraft/converter_configuration_test_data_GwGm' / 'output.dat'
        cmd = f'aviary fortran_to_aviary {filepath} -o {outfile} -l GASP'
        self.run_and_test_cmd(cmd)

    def test_small_single_aisle_conversion(self):
        filepath = pkg_resources.resource_filename('aviary',
                                                   'models/small_single_aisle/small_single_aisle_GwGm.dat')
        outfile = Path.cwd() / 'small_single_aisle' / 'output.dat'
        cmd = f'aviary fortran_to_aviary {filepath} -o {outfile} -l GASP'
        self.run_and_test_cmd(cmd)

    def test_FLOPS_conversion(self):
        filepath = pkg_resources.resource_filename('aviary',
                                                   'models/N3CC/N3CC_generic_low_speed_polars_FLOPSinp.txt')
        outfile = Path.cwd() / 'N3CC' / 'output.dat'
        cmd = f'aviary fortran_to_aviary {filepath} -o {outfile} -l FLOPS'
        self.run_and_test_cmd(cmd)

    def test_force_conversion(self):
        filepath = pkg_resources.resource_filename('aviary',
                                                   'models/test_aircraft/converter_configuration_test_data_GwGm.dat')
        outfile = Path.cwd() / 'output.dat'
        cmd1 = f'aviary fortran_to_aviary {filepath} -o {outfile} -l GASP'
        self.run_and_test_cmd(cmd1)
        filepath = pkg_resources.resource_filename('aviary',
                                                   'models/test_aircraft/converter_configuration_test_data_GwGm.dat')
        cmd2 = f'aviary fortran_to_aviary {filepath} -o {outfile} --force -l GASP'
        self.run_and_test_cmd(cmd2)


if __name__ == "__main__":
    unittest.main()



import openmdao.api as om

from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.examples.external_subsystems.battery.battery_variables import Aircraft, Mission
from aviary.examples.external_subsystems.battery.battery_variable_meta_data import ExtendedMetaData


class BatteryPreMission(om.ExplicitComponent):
    """
    Calculates battery pack mass
    """

    def setup(self):

        # Inputs
        # control from optimizer
        add_aviary_input(self, Aircraft.Battery.ENERGY_REQUIRED,
                         meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.EFFICIENCY, meta_data=ExtendedMetaData)

        # output from zappy
        add_aviary_input(self, Aircraft.Battery.CURRENT_MAX, meta_data=ExtendedMetaData)

        # params (fixed)
        add_aviary_input(self, Aircraft.Battery.Cell.HEAT_CAPACITY,
                         meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.Case.HEAT_CAPACITY,
                         meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.Cell.MASS, meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.Cell.VOLTAGE_LOW,
                         meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.Cell.DISCHARGE_RATE,
                         meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX,
                         meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.Case.WEIGHT_FRAC,
                         meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.VOLTAGE, meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.Cell.VOLUME, meta_data=ExtendedMetaData)

        add_aviary_output(self, Aircraft.Battery.N_SERIES,
                          val=0.0, meta_data=ExtendedMetaData)
        add_aviary_output(self, Aircraft.Battery.N_PARALLEL,
                          val=0.0, meta_data=ExtendedMetaData)

        add_aviary_output(self, Aircraft.Battery.MASS,
                          val=0.0, meta_data=ExtendedMetaData)
        add_aviary_output(self, Aircraft.Battery.HEAT_CAPACITY,
                          val=0.0, meta_data=ExtendedMetaData)
        # unconnected output used for test checking
        add_aviary_output(self, Aircraft.Battery.VOLUME,
                          val=0.0, meta_data=ExtendedMetaData)

        self.declare_partials(Aircraft.Battery.N_SERIES, [
                              Aircraft.Battery.VOLTAGE, Aircraft.Battery.Cell.VOLTAGE_LOW])
        self.declare_partials(Aircraft.Battery.N_PARALLEL, [
                              Aircraft.Battery.CURRENT_MAX, Aircraft.Battery.Cell.DISCHARGE_RATE])
        self.declare_partials(
            Aircraft.Battery.MASS,
            [
                Aircraft.Battery.VOLTAGE,
                Aircraft.Battery.Cell.VOLTAGE_LOW,
                Aircraft.Battery.CURRENT_MAX,
                Aircraft.Battery.Cell.DISCHARGE_RATE,
                Aircraft.Battery.Cell.MASS,
                Aircraft.Battery.Case.WEIGHT_FRAC,
            ],
        )

        self.declare_partials(Aircraft.Battery.HEAT_CAPACITY, [
                              Aircraft.Battery.Case.HEAT_CAPACITY, Aircraft.Battery.Cell.HEAT_CAPACITY, Aircraft.Battery.Case.WEIGHT_FRAC])
        self.declare_partials(
            Aircraft.Battery.VOLUME,
            [
                Aircraft.Battery.VOLTAGE,
                Aircraft.Battery.Cell.VOLTAGE_LOW,
                Aircraft.Battery.CURRENT_MAX,
                Aircraft.Battery.Cell.DISCHARGE_RATE,
                Aircraft.Battery.Cell.VOLUME,
            ],
        )

    def compute(self, inputs, outputs):

        bus_v = inputs[Aircraft.Battery.VOLTAGE]
        c_l_v = inputs[Aircraft.Battery.Cell.VOLTAGE_LOW]
        max_amps = inputs[Aircraft.Battery.CURRENT_MAX]

        outputs[Aircraft.Battery.N_SERIES] = bus_v / c_l_v
        outputs[Aircraft.Battery.N_PARALLEL] = max_amps / \
            inputs[Aircraft.Battery.Cell.DISCHARGE_RATE]
        outputs[Aircraft.Battery.MASS] = (
            outputs[Aircraft.Battery.N_SERIES]
            * outputs[Aircraft.Battery.N_PARALLEL]
            * inputs[Aircraft.Battery.Cell.MASS]
            * inputs[Aircraft.Battery.Case.WEIGHT_FRAC]
        )
        outputs[Aircraft.Battery.HEAT_CAPACITY] = (
            inputs[Aircraft.Battery.Cell.HEAT_CAPACITY] + inputs[Aircraft.Battery.Case.HEAT_CAPACITY] *
            (1.0 - inputs[Aircraft.Battery.Case.WEIGHT_FRAC])
        ) / inputs[Aircraft.Battery.Case.WEIGHT_FRAC]
        outputs[Aircraft.Battery.VOLUME] = (
            outputs[Aircraft.Battery.N_SERIES] *
            outputs[Aircraft.Battery.N_PARALLEL] * inputs[Aircraft.Battery.Cell.VOLUME]
        )

    def compute_partials(self, inputs, partials):

        bus_v = inputs[Aircraft.Battery.VOLTAGE]
        c_l_v = inputs[Aircraft.Battery.Cell.VOLTAGE_LOW]
        max_amps = inputs[Aircraft.Battery.CURRENT_MAX]
        I_rate = inputs[Aircraft.Battery.Cell.DISCHARGE_RATE]
        m_cell = inputs[Aircraft.Battery.Cell.MASS]
        wf_case = inputs[Aircraft.Battery.Case.WEIGHT_FRAC]
        n_s = bus_v / c_l_v
        n_p = max_amps / inputs[Aircraft.Battery.Cell.DISCHARGE_RATE]
        cp_case = inputs[Aircraft.Battery.Case.HEAT_CAPACITY]
        cp_cell = inputs[Aircraft.Battery.Cell.HEAT_CAPACITY]
        vc = inputs[Aircraft.Battery.Cell.VOLUME]

        # n_{series}
        partials[Aircraft.Battery.N_SERIES, Aircraft.Battery.VOLTAGE] = 1.0 / c_l_v
        partials[Aircraft.Battery.N_SERIES,
                 Aircraft.Battery.Cell.VOLTAGE_LOW] = -bus_v / (c_l_v ** 2)
        # n_{parallel}
        partials[Aircraft.Battery.N_PARALLEL,
                 Aircraft.Battery.CURRENT_MAX] = 1.0 / I_rate
        partials[Aircraft.Battery.N_PARALLEL,
                 Aircraft.Battery.Cell.DISCHARGE_RATE] = -max_amps / (I_rate ** 2)
        # mass_{battery}
        partials[Aircraft.Battery.MASS, Aircraft.Battery.VOLTAGE] = (
            partials[Aircraft.Battery.N_SERIES,
                     Aircraft.Battery.VOLTAGE] * n_p * m_cell * wf_case
        )
        partials[Aircraft.Battery.MASS, Aircraft.Battery.Cell.VOLTAGE_LOW] = (
            partials[Aircraft.Battery.N_SERIES,
                     Aircraft.Battery.Cell.VOLTAGE_LOW] * n_p * m_cell * wf_case
        )
        partials[Aircraft.Battery.MASS, Aircraft.Battery.CURRENT_MAX] = (
            partials[Aircraft.Battery.N_PARALLEL,
                     Aircraft.Battery.CURRENT_MAX] * n_s * m_cell * wf_case
        )
        partials[Aircraft.Battery.MASS, Aircraft.Battery.Cell.DISCHARGE_RATE] = (
            partials[Aircraft.Battery.N_PARALLEL,
                     Aircraft.Battery.Cell.DISCHARGE_RATE] * n_s * m_cell * wf_case
        )
        partials[Aircraft.Battery.MASS, Aircraft.Battery.Cell.MASS] = n_s * n_p * wf_case
        partials[Aircraft.Battery.MASS,
                 Aircraft.Battery.Case.WEIGHT_FRAC] = n_s * n_p * m_cell
        # C_{p,batt}
        partials[Aircraft.Battery.HEAT_CAPACITY,
                 Aircraft.Battery.Cell.HEAT_CAPACITY] = 1.0 / wf_case
        partials[Aircraft.Battery.HEAT_CAPACITY,
                 Aircraft.Battery.Case.HEAT_CAPACITY] = (1.0 - wf_case) / wf_case
        partials[Aircraft.Battery.HEAT_CAPACITY, Aircraft.Battery.Case.WEIGHT_FRAC] = (
            -(cp_cell + cp_case) / wf_case ** 2
        )
        # volume_{pack}
        partials[Aircraft.Battery.VOLUME, Aircraft.Battery.VOLTAGE] = (
            partials[Aircraft.Battery.N_SERIES, Aircraft.Battery.VOLTAGE] * n_p * vc
        )
        partials[Aircraft.Battery.VOLUME, Aircraft.Battery.Cell.VOLTAGE_LOW] = (
            partials[Aircraft.Battery.N_SERIES,
                     Aircraft.Battery.Cell.VOLTAGE_LOW] * n_p * vc
        )
        partials[Aircraft.Battery.VOLUME, Aircraft.Battery.CURRENT_MAX] = (
            partials[Aircraft.Battery.N_PARALLEL,
                     Aircraft.Battery.CURRENT_MAX] * n_s * vc
        )
        partials[Aircraft.Battery.VOLUME, Aircraft.Battery.Cell.DISCHARGE_RATE] = (
            partials[Aircraft.Battery.N_PARALLEL,
                     Aircraft.Battery.Cell.DISCHARGE_RATE] * n_s * vc
        )
        partials[Aircraft.Battery.VOLUME, Aircraft.Battery.Cell.VOLUME] = n_s * n_p



"""
Define subsystem builder for Aviary core propulsion.

Classes
-------
PropulsionBuilderBase : the interface for a propulsion subsystem builder.

CorePropulsionBuilder : the interface for Aviary's core propulsion subsystem builder
"""
from aviary.interface.utils.markdown_utils import write_markdown_variable_table
from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.subsystems.propulsion.propulsion_premission import PropulsionPreMission
from aviary.subsystems.propulsion.propulsion_mission import PropulsionMission
from aviary.variable_info.variables import Aircraft

_default_name = 'propulsion'


class PropulsionBuilderBase(SubsystemBuilderBase):
    def __init__(self, name=None, meta_data=None):
        if name is None:
            name = _default_name

        super().__init__(name=name, meta_data=meta_data)

    def mission_inputs(self, **kwargs):
        return ['*']

    def mission_outputs(self, **kwargs):
        return ['*']


class CorePropulsionBuilder(PropulsionBuilderBase):
    # code_origin is not necessary for this subsystem, catch with kwargs and ignore
    def __init__(self, name=None, meta_data=None, **kwargs):
        if name is None:
            name = 'core_propulsion'

        super().__init__(name=name, meta_data=meta_data)

    def build_pre_mission(self, aviary_inputs):
        return PropulsionPreMission(aviary_options=aviary_inputs)

    def build_mission(self, num_nodes, aviary_inputs, **kwargs):
        return PropulsionMission(num_nodes=num_nodes, aviary_options=aviary_inputs)

    def report(self, prob, reports_folder, **kwargs):
        """
        Generate the report for Aviary core propulsion analysis

        Parameters
        ----------
        prob : AviaryProblem
            The AviaryProblem that will be used to generate the report
        reports_folder : Path
            Location of the subsystems_report folder this report will be placed in
        """
        filename = self.name + '.md'
        filepath = reports_folder / filename

        propulsion_outputs = [Aircraft.Propulsion.TOTAL_NUM_ENGINES,
                              Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST]

        engine_outputs = [Aircraft.Engine.NUM_ENGINES,
                          Aircraft.Engine.SCALE_FACTOR,
                          Aircraft.Engine.SCALED_SLS_THRUST]

        with open(filepath, mode='w') as f:
            f.write('# PROPULSION')
            write_markdown_variable_table(f, prob, propulsion_outputs, self.meta_data)
            f.write('\n## ENGINES')
            for engine in prob.aviary_inputs.get_val('engine_models'):
                f.write(f'\n### {engine.name}')
                write_markdown_variable_table(f, engine, engine_outputs, self.meta_data)



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class VerticalTailMass(om.ExplicitComponent):
    '''
    Calculates the mass of the vertical tail(s). The methodology is based on the FLOPS weight
    equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.VerticalTail.AREA, val=0.0)

        add_aviary_input(self, Aircraft.VerticalTail.TAPER_RATIO, val=0.0)

        add_aviary_input(self, Aircraft.VerticalTail.MASS_SCALER, val=1.0)

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)

        add_aviary_output(self, Aircraft.VerticalTail.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_tails = aviary_options.get_val(Aircraft.VerticalTail.NUM_TAILS)

        area = inputs[Aircraft.VerticalTail.AREA]
        taper_ratio = inputs[Aircraft.VerticalTail.TAPER_RATIO]
        scaler = inputs[Aircraft.VerticalTail.MASS_SCALER]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM

        outputs[Aircraft.VerticalTail.MASS] = scaler * 0.32 * \
            gross_weight ** 0.30 * (taper_ratio + 0.50) * area ** 0.85 * \
            num_tails ** 0.7 / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_tails = aviary_options.get_val(Aircraft.VerticalTail.NUM_TAILS)

        area = inputs[Aircraft.VerticalTail.AREA]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        taper_ratio = inputs[Aircraft.VerticalTail.TAPER_RATIO]
        scaler = inputs[Aircraft.VerticalTail.MASS_SCALER]

        gross_weight_exp = gross_weight ** 0.30
        area_exp = area ** 0.85
        num_tails_exp = num_tails ** 0.7

        J[Aircraft.VerticalTail.MASS, Aircraft.VerticalTail.AREA] = \
            scaler * 0.272 * gross_weight_exp * \
            (taper_ratio + 0.50) * area ** -0.15 * \
            num_tails_exp / GRAV_ENGLISH_LBM

        J[Aircraft.VerticalTail.MASS, Mission.Design.GROSS_MASS] = \
            scaler * 0.096 * gross_weight ** -0.70 * \
            (taper_ratio + 0.50) * area_exp * num_tails_exp

        J[Aircraft.VerticalTail.MASS, Aircraft.VerticalTail.TAPER_RATIO] = \
            scaler * 0.32 * gross_weight_exp * area_exp * num_tails_exp / GRAV_ENGLISH_LBM

        J[Aircraft.VerticalTail.MASS, Aircraft.VerticalTail.MASS_SCALER] = \
            0.32 * gross_weight_exp * (taper_ratio + 0.50) * \
            area_exp * num_tails_exp / GRAV_ENGLISH_LBM


class AltVerticalTailMass(om.ExplicitComponent):
    '''
    Calculates the mass of the vertical tail(s) using the alternate method.
    The methodology is based on the FLOPS weight- equations, modified to
    output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.VerticalTail.AREA, val=0.0)

        add_aviary_input(self, Aircraft.VerticalTail.MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.VerticalTail.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        area = inputs[Aircraft.VerticalTail.AREA]
        scaler = inputs[Aircraft.VerticalTail.MASS_SCALER]

        outputs[Aircraft.VerticalTail.MASS] = scaler * 6.0 * area

    def compute_partials(self, inputs, J):
        area = inputs[Aircraft.VerticalTail.AREA]
        scaler = inputs[Aircraft.VerticalTail.MASS_SCALER]

        J[Aircraft.VerticalTail.MASS, Aircraft.VerticalTail.AREA] = (
            6.0 * scaler
        )

        J[Aircraft.VerticalTail.MASS, Aircraft.VerticalTail.MASS_SCALER] = (
            6.0 * area
        )



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.subsystems.aerodynamics.gasp_based.flaps_model.Cl_max import \
    CLmaxCalculation
from aviary.variable_info.variables import Aircraft, Dynamic

"""
All data is from validation files using standalone flaps model
"""


class CLmaxCalculationTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem(model=om.Group())

        self.prob.model.add_subsystem('CLmC', CLmaxCalculation(), promotes=['*'])

        self.prob.setup()

        # initial conditions
        self.prob.set_val("VLAM1", 0.97217)
        self.prob.set_val("VLAM2", 1.09948)
        self.prob.set_val("VLAM3", 0.97217)
        self.prob.set_val("VLAM4", 1.25725)
        self.prob.set_val("VLAM5", 1.0000)
        self.prob.set_val("VLAM6", 1.0000)
        self.prob.set_val("VLAM7", 0.735)
        self.prob.set_val("VLAM8", 0.74444)
        self.prob.set_val("VLAM9", 0.9975)
        self.prob.set_val("VLAM10", 0.74)
        self.prob.set_val("VLAM11", 0.84232)
        self.prob.set_val("VLAM12", 0.79208)
        self.prob.set_val("VLAM13", 1.03512)
        self.prob.set_val("VLAM14", 0.99124)

        self.prob.set_val(Dynamic.Mission.SPEED_OF_SOUND, 1118.21948771, units="ft/s")  #
        self.prob.set_val(Aircraft.Wing.LOADING, 128.0, units="lbf/ft**2")
        self.prob.set_val(Dynamic.Mission.STATIC_PRESSURE,
                          (14.696 * 144), units="lbf/ft**2")
        self.prob.set_val(Aircraft.Wing.AVERAGE_CHORD, 12.61, units="ft")
        self.prob.set_val("kinematic_viscosity", 0.15723e-3, units="ft**2/s")
        self.prob.set_val(Aircraft.Wing.MAX_LIFT_REF, 1.150)
        self.prob.set_val(Aircraft.Wing.SLAT_LIFT_INCREMENT_OPTIMUM, 0.930)
        self.prob.set_val("fus_lift", 0.05498)
        self.prob.set_val(Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM, 1.500)
        self.prob.set_val(Dynamic.Mission.TEMPERATURE, 518.7, units="degR")

    def test_case(self):

        self.prob.run_model()
        tol = 6e-4
        print()

        reg_data = 2.8155
        ans = self.prob["CL_max"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.17522
        ans = self.prob[Dynamic.Mission.MACH]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 157.19864
        ans = self.prob["reynolds"]
        assert_near_equal(ans, reg_data, tol)

        data = self.prob.check_partials(out_stream=None, method="fd")
        assert_check_partials(
            data, atol=6400, rtol=0.007
        )  # large to account for large magnitude value, discrepancies are acceptable


if __name__ == "__main__":
    unittest.main()



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.avionics import TransportAvionicsMass
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Mission

_fuse_count = 'TBD:option:fuselage_count'
_fuse_total_length = 'aircraft:fuselage:dimensions:total_length'
_fuse_max_depth = 'aircraft:fuselage:dimensions:max_depth'
_max_mach = 'aircraft:design:dimensions:max_mach'
# _carrier_based = 'aircraft:landing_gear:control:carrier_based'
_avionics_group_mass = 'TBD:avionics'


class TransportAvionicsMassTest(unittest.TestCase):
    '''
    Tests transport/GA avionics mass calculation.
    '''

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            "avionics",
            TransportAvionicsMass(aviary_options=get_flops_inputs(
                case_name, preprocess=True)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Avionics.MASS_SCALER,
                        Aircraft.Fuselage.PLANFORM_AREA,
                        Mission.Design.RANGE],
            output_keys=Aircraft.Avionics.MASS,
            aviary_option_keys=[Aircraft.CrewPayload.NUM_FLIGHT_CREW],
            tol=2.0e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == '__main__':
    unittest.main()



import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class FuelCapacityGroup(om.Group):
    """
    Compute the maximum fuel that can be carried.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        aviary_options = self.options['aviary_options']

        self.add_subsystem(
            'wing_fuel_capacity',
            WingFuelCapacity(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'fuselage_fuel_capacity',
            FuselageFuelCapacity(),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'auxiliary_fuel_capacity',
            AuxFuelCapacity(),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'total_fuel_capacity',
            TotalFuelCapacity(),
            promotes_inputs=['*'], promotes_outputs=['*'])


class FuselageFuelCapacity(om.ExplicitComponent):
    """
    Compute the maximum fuel that can be carried in the fuselage.
    """

    def setup(self):
        add_aviary_input(self, Aircraft.Fuel.TOTAL_CAPACITY, 0.0),
        add_aviary_input(self, Aircraft.Fuel.WING_FUEL_CAPACITY, 0.0),
        add_aviary_output(self, Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY, 0.0),

    def setup_partials(self):
        self.declare_partials(
            Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY,
            Aircraft.Fuel.TOTAL_CAPACITY, val=1.0)
        self.declare_partials(
            Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY,
            Aircraft.Fuel.WING_FUEL_CAPACITY, val=-1.0)

    def compute(self, inputs, outputs):
        outputs[Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY] = (
            inputs[Aircraft.Fuel.TOTAL_CAPACITY]
            - inputs[Aircraft.Fuel.WING_FUEL_CAPACITY])


class AuxFuelCapacity(om.ExplicitComponent):
    """
    Compute the maximum fuel that can be carried in the auxiliary tanks.
    """

    def setup(self):
        add_aviary_input(self, Aircraft.Fuel.TOTAL_CAPACITY, 0.0),
        add_aviary_input(self, Aircraft.Fuel.WING_FUEL_CAPACITY, 0.0),
        add_aviary_input(self, Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY, 0.0),
        add_aviary_output(self, Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY, 0.0),

    def setup_partials(self):
        self.declare_partials(
            Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
            Aircraft.Fuel.TOTAL_CAPACITY, val=1.0)
        self.declare_partials(
            Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
            Aircraft.Fuel.WING_FUEL_CAPACITY, val=-1.0)
        self.declare_partials(
            Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
            Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY, val=-1.0)

    def compute(self, inputs, outputs):
        outputs[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY] = (
            inputs[Aircraft.Fuel.TOTAL_CAPACITY]
            - inputs[Aircraft.Fuel.WING_FUEL_CAPACITY]
            - inputs[Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY])


class TotalFuelCapacity(om.ExplicitComponent):
    """
    Compute the total fuel that can be carried in all tanks.
    """

    def setup(self):
        add_aviary_input(self, Aircraft.Fuel.WING_FUEL_CAPACITY, 0.0),
        add_aviary_input(self, Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY, 0.0),
        add_aviary_input(self, Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY, 0.0),
        add_aviary_output(self, Aircraft.Fuel.TOTAL_CAPACITY, 0.0),

    def setup_partials(self):
        self.declare_partials(
            Aircraft.Fuel.TOTAL_CAPACITY,
            Aircraft.Fuel.WING_FUEL_CAPACITY, val=1.0)
        self.declare_partials(
            Aircraft.Fuel.TOTAL_CAPACITY,
            Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY, val=1.0)
        self.declare_partials(
            Aircraft.Fuel.TOTAL_CAPACITY,
            Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY, val=1.0)

    def compute(self, inputs, outputs):
        outputs[Aircraft.Fuel.TOTAL_CAPACITY] = (
            inputs[Aircraft.Fuel.WING_FUEL_CAPACITY]
            + inputs[Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY]
            + inputs[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY])


class WingFuelCapacity(om.ExplicitComponent):
    """
    Compute the maximum fuel that can be carried in the wing's enclosed space.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuel.DENSITY_RATIO, 1.0)
        add_aviary_input(self, Aircraft.Fuel.WING_REF_CAPACITY, 0.0)
        add_aviary_input(self, Aircraft.Fuel.WING_REF_CAPACITY_AREA, 0.0)
        add_aviary_input(self, Aircraft.Fuel.WING_REF_CAPACITY_TERM_A, 0.0)
        add_aviary_input(self, Aircraft.Fuel.WING_REF_CAPACITY_TERM_B, 0.0)

        add_aviary_input(self, Aircraft.Fuel.CAPACITY_FACTOR, 23.0)

        add_aviary_input(self, Aircraft.Wing.AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.SPAN, 0.0)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, 0.0)

        add_aviary_output(self, Aircraft.Fuel.WING_FUEL_CAPACITY, 0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        wing_ref_cap_terma = inputs[Aircraft.Fuel.WING_REF_CAPACITY_TERM_A]
        wing_area = inputs[Aircraft.Wing.AREA]

        if wing_ref_cap_terma.real > 0.0:
            wing_ref_cap = inputs[Aircraft.Fuel.WING_REF_CAPACITY]
            wing_ref_cap_area = inputs[Aircraft.Fuel.WING_REF_CAPACITY_AREA]
            wing_ref_cap_termb = inputs[Aircraft.Fuel.WING_REF_CAPACITY_TERM_B]

            fuel_cap_wing = (
                wing_ref_cap
                + wing_ref_cap_terma * (wing_area**1.5 - wing_ref_cap_area**1.5)
                + wing_ref_cap_termb * (wing_area - wing_ref_cap_area))

        else:
            density_ratio = inputs[Aircraft.Fuel.DENSITY_RATIO]
            cap_factor = inputs[Aircraft.Fuel.CAPACITY_FACTOR]
            span = inputs[Aircraft.Wing.SPAN]
            taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
            thickness_to_chord = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]

            fuel_cap_wing = (
                density_ratio * cap_factor * wing_area**2 * thickness_to_chord
                * (1.0 - taper_ratio / (1.0 + taper_ratio)**2) / span)

        outputs[Aircraft.Fuel.WING_FUEL_CAPACITY] = fuel_cap_wing

    def compute_partials(self, inputs, partials):
        wing_ref_cap_terma = inputs[Aircraft.Fuel.WING_REF_CAPACITY_TERM_A]
        wing_area = inputs[Aircraft.Wing.AREA]

        if wing_ref_cap_terma.real > 0.0:
            wing_ref_cap_area = inputs[Aircraft.Fuel.WING_REF_CAPACITY_AREA]
            wing_ref_cap_termb = inputs[Aircraft.Fuel.WING_REF_CAPACITY_TERM_B]

            partials[
                Aircraft.Fuel.WING_FUEL_CAPACITY, Aircraft.Fuel.WING_REF_CAPACITY] = 1.0

            partials[
                Aircraft.Fuel.WING_FUEL_CAPACITY,
                Aircraft.Fuel.WING_REF_CAPACITY_TERM_A] = \
                wing_area**1.5 - wing_ref_cap_area**1.5

            partials[
                Aircraft.Fuel.WING_FUEL_CAPACITY,
                Aircraft.Fuel.WING_REF_CAPACITY_TERM_B] = wing_area - wing_ref_cap_area

            partials[Aircraft.Fuel.WING_FUEL_CAPACITY, Aircraft.Wing.AREA] = \
                1.5 * wing_ref_cap_terma * wing_area**0.5 + wing_ref_cap_termb

            partials[
                Aircraft.Fuel.WING_FUEL_CAPACITY,
                Aircraft.Fuel.WING_REF_CAPACITY_AREA] = \
                -1.5 * wing_ref_cap_terma * wing_ref_cap_area**0.5 - wing_ref_cap_termb

        else:
            density_ratio = inputs[Aircraft.Fuel.DENSITY_RATIO]
            cap_factor = inputs[Aircraft.Fuel.CAPACITY_FACTOR]
            span = inputs[Aircraft.Wing.SPAN]
            taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
            thickness_to_chord = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]

            den = (1.0 + taper_ratio)
            tr_fact = (1.0 - taper_ratio / den**2)
            dfact = - 1.0/den**2 + 2.0 * taper_ratio / den**3

            partials[Aircraft.Fuel.WING_FUEL_CAPACITY, Aircraft.Fuel.DENSITY_RATIO] = \
                cap_factor * wing_area**2 * thickness_to_chord * tr_fact / span

            partials[Aircraft.Fuel.WING_FUEL_CAPACITY, Aircraft.Fuel.CAPACITY_FACTOR] = \
                density_ratio * wing_area**2 * thickness_to_chord * tr_fact / span

            partials[Aircraft.Fuel.WING_FUEL_CAPACITY, Aircraft.Wing.SPAN] = (
                -density_ratio * cap_factor * wing_area**2 * thickness_to_chord * tr_fact
                / span**2)

            partials[Aircraft.Fuel.WING_FUEL_CAPACITY, Aircraft.Wing.TAPER_RATIO] = (
                density_ratio * cap_factor * wing_area**2 * thickness_to_chord * dfact
                / span)

            partials[
                Aircraft.Fuel.WING_FUEL_CAPACITY, Aircraft.Wing.THICKNESS_TO_CHORD] = \
                density_ratio * cap_factor * wing_area**2 * tr_fact / span

            partials[Aircraft.Fuel.WING_FUEL_CAPACITY, Aircraft.Wing.AREA] = (
                2.0 * density_ratio * cap_factor * wing_area * thickness_to_chord
                * tr_fact / span)



import csv
import unittest
from pathlib import Path

from openmdao.utils.assert_utils import assert_near_equal

from aviary.subsystems.propulsion.engine_deck import EngineDeck
from aviary.subsystems.propulsion.utils import EngineModelVariables as keys
from aviary.utils.named_values import NamedValues
from aviary.validation_cases.validation_data.flops_data.FLOPS_Test_Data import \
    FLOPS_Test_Data


class EngineDeckTest(unittest.TestCase):
    def test_flight_idle(self):
        tol = 1e-6

        aviary_values = FLOPS_Test_Data['LargeSingleAisle2FLOPS']['inputs']

        model = aviary_values.get_val('engine_models')[0]

        expected_mach_number = []
        expected_altitude = []
        expected_throttle = []
        expected_thrust = []
        expected_fuel_flow_rate = []

        # hardcoded data of processed engine model from LEAPS1 after flight idle
        # point generation, sorted in Aviary order
        with open(Path(__file__).parents[0] / 'engine_model_test_data_turbofan_24k_1.csv') as file:
            reader = csv.reader(file)
            for row in reader:
                expected_mach_number.append(float(row[0]))
                expected_altitude.append(float(row[1]))
                expected_throttle.append(float(row[2]))
                expected_thrust.append(float(row[3]))
                expected_fuel_flow_rate.append(float(row[4]))

        mach_number = model.data[keys.MACH]
        altitude = model.data[keys.ALTITUDE]
        throttle = model.data[keys.THROTTLE]
        thrust = model.data[keys.THRUST]
        fuel_flow_rate = model.data[keys.FUEL_FLOW]

        assert_near_equal(mach_number, expected_mach_number, tolerance=tol)
        assert_near_equal(altitude, expected_altitude, tolerance=tol)
        # assert_near_equal(throttle, expected_throttle, tolerance=tol)
        assert_near_equal(thrust, expected_thrust, tolerance=tol)
        assert_near_equal(fuel_flow_rate, expected_fuel_flow_rate, tolerance=tol)

    def test_flight_idle_2(self):
        tol = 1e-6

        aviary_values = FLOPS_Test_Data['LargeSingleAisle1FLOPS']['inputs']

        model = aviary_values.get_val('engine_models')[0]

        # hardcoded data of processed engine model from LEAPS1 after flight idle
        # point generation, sorted in Aviary order

        expected_mach_number = []
        expected_altitude = []
        expected_throttle = []
        expected_thrust = []
        expected_fuel_flow_rate = []

        # hardcoded data of processed engine model from LEAPS1 after flight idle
        # point generation, sorted in Aviary order
        with open(Path(__file__).parents[0] / 'engine_model_test_data_turbofan_28k.csv') as file:
            reader = csv.reader(file)
            for row in reader:
                expected_mach_number.append(float(row[0]))
                expected_altitude.append(float(row[1]))
                expected_throttle.append(float(row[2]))
                expected_thrust.append(float(row[3]))
                expected_fuel_flow_rate.append(float(row[4]))

        mach_number = model.data[keys.MACH]
        altitude = model.data[keys.ALTITUDE]
        throttle = model.data[keys.THROTTLE]
        thrust = model.data[keys.THRUST]
        fuel_flow_rate = model.data[keys.FUEL_FLOW]

        assert_near_equal(mach_number, expected_mach_number, tolerance=tol)
        assert_near_equal(altitude, expected_altitude, tolerance=tol)
        # assert_near_equal(throttle, expected_throttle, tolerance=tol)
        assert_near_equal(thrust, expected_thrust, tolerance=tol)
        assert_near_equal(fuel_flow_rate, expected_fuel_flow_rate, tolerance=tol)

    def test_load_from_memory(self):
        tol = 1e-6

        aviary_values = FLOPS_Test_Data['LargeSingleAisle2FLOPS']['inputs']

        expected_mach_number = []
        expected_altitude = []
        expected_throttle = []
        expected_thrust = []
        expected_fuel_flow_rate = []

        # hardcoded data of processed engine model from LEAPS1 after flight idle
        # point generation, sorted in Aviary order
        with open(Path(__file__).parents[0] / 'engine_model_test_data_turbofan_24k_1.csv') as file:
            reader = csv.reader(file)
            for row in reader:
                expected_mach_number.append(float(row[0]))
                expected_altitude.append(float(row[1]))
                expected_throttle.append(float(row[2]))
                expected_thrust.append(float(row[3]))
                expected_fuel_flow_rate.append(float(row[4]))

        data_input = NamedValues()
        data_input.set_val('mach', expected_mach_number, 'unitless')
        data_input.set_val('altitude', expected_altitude, 'ft')
        data_input.set_val('throttle', expected_throttle, 'unitless')
        data_input.set_val('thrust', expected_thrust, 'lbf')
        data_input.set_val('fuel_flow', expected_fuel_flow_rate, 'lbm/h')

        model = EngineDeck('engine', aviary_values, data_input)

        mach_number = model.data[keys.MACH]
        altitude = model.data[keys.ALTITUDE]
        throttle = model.data[keys.THROTTLE]
        thrust = model.data[keys.THRUST]
        fuel_flow_rate = model.data[keys.FUEL_FLOW]

        assert_near_equal(mach_number, expected_mach_number, tolerance=tol)
        assert_near_equal(altitude, expected_altitude, tolerance=tol)
        # assert_near_equal(throttle, expected_throttle, tolerance=tol)
        assert_near_equal(thrust, expected_thrust, tolerance=tol)
        assert_near_equal(fuel_flow_rate, expected_fuel_flow_rate, tolerance=tol)


if __name__ == "__main__":
    # unittest.main()
    test = EngineDeckTest()
    test.test_load_from_memory()



from aviary.mission.phase_builder_base import PhaseBuilderBase
from aviary.mission.initial_guess_builders import InitialGuessState, InitialGuessTime, InitialGuessControl
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Dynamic
from aviary.variable_info.enums import SpeedType
from aviary.mission.gasp_based.ode.descent_ode import DescentODE


class DescentPhase(PhaseBuilderBase):
    default_name = 'descent_phase'
    default_ode_class = DescentODE

    _meta_data_ = {}
    _initial_guesses_meta_data_ = {}

    def build_phase(self, aviary_options: AviaryValues = None):
        phase = self.phase = super().build_phase(aviary_options)

        # Retrieve user options values
        user_options = self.user_options
        input_speed_type = user_options.get_val('input_speed_type')
        EAS_limit = user_options.get_val('EAS_limit', units='kn')

        # Add states
        self.add_altitude_state(user_options)

        self.add_mass_state(user_options)

        self.add_distance_state(user_options)

        # Add boundary constraint
        self.add_altitude_constraint(user_options)

        # Add parameter if necessary
        if input_speed_type == SpeedType.EAS:
            phase.add_parameter("EAS", opt=False, units="kn", val=EAS_limit)

        # Add timeseries outputs
        phase.add_timeseries_output(
            Dynamic.Mission.MACH, output_name=Dynamic.Mission.MACH, units="unitless")
        phase.add_timeseries_output("EAS", output_name="EAS", units="kn")
        phase.add_timeseries_output("TAS", output_name="TAS", units="kn")
        phase.add_timeseries_output(Dynamic.Mission.FLIGHT_PATH_ANGLE,
                                    output_name=Dynamic.Mission.FLIGHT_PATH_ANGLE, units="deg")
        phase.add_timeseries_output("alpha", output_name="alpha", units="deg")
        phase.add_timeseries_output("theta", output_name="theta", units="deg")
        phase.add_timeseries_output("aero.CL", output_name="CL", units="unitless")
        phase.add_timeseries_output(
            Dynamic.Mission.THRUST_TOTAL, output_name=Dynamic.Mission.THRUST_TOTAL, units="lbf")
        phase.add_timeseries_output("aero.CD", output_name="CD", units="unitless")

        return phase

    def _extra_ode_init_kwargs(self):
        """
        Return extra kwargs required for initializing the ODE.
        """
        # TODO: support external_subsystems and meta_data in the base class
        return {
            'input_speed_type': self.user_options.get_val('input_speed_type'),
            'mach_cruise': self.user_options.get_val('mach_cruise'),
            'EAS_limit': self.user_options.get_val('EAS_limit', 'kn'),
        }


# Adding metadata for the DescentPhase
DescentPhase._add_meta_data('fix_initial', val=False)
DescentPhase._add_meta_data('input_initial', val=False)
DescentPhase._add_meta_data('EAS_limit', val=0, units='kn')
DescentPhase._add_meta_data('mach_cruise', val=0)
DescentPhase._add_meta_data('input_speed_type', val=SpeedType.MACH)
DescentPhase._add_meta_data('final_altitude', val=0, units='ft')
DescentPhase._add_meta_data('duration_bounds', val=(0, 0), units='s')
DescentPhase._add_meta_data('duration_ref', val=1, units='s')
DescentPhase._add_meta_data('alt_lower', val=0, units='ft')
DescentPhase._add_meta_data('alt_upper', val=0, units='ft')
DescentPhase._add_meta_data('alt_ref', val=1, units='ft')
DescentPhase._add_meta_data('alt_ref0', val=0, units='ft')
DescentPhase._add_meta_data('alt_defect_ref', val=None, units='ft')
DescentPhase._add_meta_data('alt_constraint_ref', val=None, units='ft')
DescentPhase._add_meta_data('mass_lower', val=0, units='lbm')
DescentPhase._add_meta_data('mass_upper', val=0, units='lbm')
DescentPhase._add_meta_data('mass_ref', val=1, units='lbm')
DescentPhase._add_meta_data('mass_ref0', val=0, units='lbm')
DescentPhase._add_meta_data('mass_defect_ref', val=None, units='lbm')
DescentPhase._add_meta_data('distance_lower', val=0, units='NM')
DescentPhase._add_meta_data('distance_upper', val=0, units='NM')
DescentPhase._add_meta_data('distance_ref', val=1, units='NM')
DescentPhase._add_meta_data('distance_ref0', val=0, units='NM')
DescentPhase._add_meta_data('distance_defect_ref', val=None, units='NM')
DescentPhase._add_meta_data('num_segments', val=None, units='unitless')
DescentPhase._add_meta_data('order', val=None, units='unitless')

# Adding initial guess metadata
DescentPhase._add_initial_guess_meta_data(
    InitialGuessTime(),
    desc='initial guess for time options')
DescentPhase._add_initial_guess_meta_data(
    InitialGuessState('altitude'),
    desc='initial guess for altitude state')
DescentPhase._add_initial_guess_meta_data(
    InitialGuessState('mass'),
    desc='initial guess for mass state')
DescentPhase._add_initial_guess_meta_data(
    InitialGuessState('distance'),
    desc='initial guess for distance state')
DescentPhase._add_initial_guess_meta_data(
    InitialGuessControl('throttle'),
    desc='initial guess for throttle')



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission

_fuse_count = 'TBD:option:fuselage_count'
_fuse_total_length = 'aircraft:fuselage:dimensions:total_length'
_fuse_max_depth = 'aircraft:fuselage:dimensions:max_depth'
_max_mach = 'aircraft:design:dimensions:max_mach'
_carrier_based = 'carrier_based'  # TODO: options can't use : in the name
_avionics_group_mass = 'TBD:avionics'


class TransportAvionicsMass(om.ExplicitComponent):
    '''
    Calculates the mass of the avionics group using the transport/general aviation method.
    The methodology is based on the FLOPS weight equations, modified to output mass
    instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Avionics.MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.Fuselage.PLANFORM_AREA, val=0.0)

        add_aviary_input(self, Mission.Design.RANGE, val=0.0)

        add_aviary_output(self, Aircraft.Avionics.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        crew = aviary_options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_CREW)
        scaler = inputs[Aircraft.Avionics.MASS_SCALER]
        planform = inputs[Aircraft.Fuselage.PLANFORM_AREA]
        des_range = inputs[Mission.Design.RANGE]

        outputs[Aircraft.Avionics.MASS] = \
            15.8 * des_range**0.1 * crew**0.7 * planform**0.43 * scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        crew = aviary_options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_CREW)
        scaler = inputs[Aircraft.Avionics.MASS_SCALER]
        planform = inputs[Aircraft.Fuselage.PLANFORM_AREA]
        des_range = inputs[Mission.Design.RANGE]

        des_range_exp = des_range**0.1
        crew_exp = crew**0.7
        planform_exp = planform**0.43

        J[Aircraft.Avionics.MASS, Aircraft.Avionics.MASS_SCALER] = 15.8 * \
            des_range_exp * crew_exp * planform_exp / GRAV_ENGLISH_LBM

        J[Aircraft.Avionics.MASS, Aircraft.Fuselage.PLANFORM_AREA] = 6.794 * \
            des_range_exp * crew_exp * planform**-0.57 * scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Avionics.MASS, Mission.Design.RANGE] = 1.58 * \
            des_range**-0.9 * crew_exp * planform_exp * scaler / GRAV_ENGLISH_LBM



import numpy as np
import openmdao.api as om

from aviary.variable_info.variables import Dynamic


class SpeedConstraints(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare(
            "EAS_target",
            default=0,
            desc="targeted equivalent airspeed in knots assuming mach constraint is"
            " satisfied",
        )
        self.options.declare(
            "mach_cruise", default=0, desc="targeted cruise mach number"
        )

    def setup(self):
        nn = self.options["num_nodes"]
        arange = np.arange(nn)

        self.add_input(
            "EAS",
            val=np.ones(nn),
            units="kn",
            desc="equivalent airspeed",
        )
        self.add_input(
            Dynamic.Mission.MACH,
            val=np.ones(nn),
            units="unitless",
            desc="mach number",
        )

        self.add_output(
            "speed_constraint",
            val=np.ones((nn, 2)),
            units="unitless",
            desc="constraint to be driven to zero in order to control speed",
        )

        self.declare_partials(
            "speed_constraint", "EAS", rows=arange * 2, cols=arange, val=1.0
        )
        self.declare_partials(
            "speed_constraint",
            Dynamic.Mission.MACH,
            rows=arange * 2 + 1,
            cols=arange,
            val=self.options["EAS_target"],
        )

    def compute(self, inputs, outputs):

        EAS = inputs["EAS"]
        EAS_target = self.options["EAS_target"]
        mach = inputs[Dynamic.Mission.MACH]
        mach_cruise = self.options["mach_cruise"]

        EAS_constraint = EAS - EAS_target
        EAS_constraint = EAS_constraint[:, np.newaxis]
        mach_constraint = EAS_target * (mach - mach_cruise)
        mach_constraint = mach_constraint[:, np.newaxis]
        outputs["speed_constraint"] = np.hstack((EAS_constraint, mach_constraint))



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM, RHO_SEA_LEVEL_ENGLISH
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic


class VRotateComp(om.ExplicitComponent):
    """
    Component that computes V_rotate based on vehicle properties and speed buffers.
    """

    def setup(self):
        # Temporarily set this to shape (1, 1) to avoid OpenMDAO bug
        self.add_input(Dynamic.Mission.MASS, shape=(1, 1), units="lbm")
        self.add_input("rho", shape=(1,), units="slug/ft**3",
                       val=RHO_SEA_LEVEL_ENGLISH, desc="sea-level atmospheric density")
        add_aviary_input(self, Aircraft.Wing.AREA, val=1.0)
        self.add_input("CL_max", shape=(1,), units="unitless",
                       desc="Maximum lift coefficient")
        self.add_input("dV1", shape=(1,), units="ft/s",
                       desc="Increment of engine failure decision speed above stall speed.")
        self.add_input("dVR", shape=(1,), units="ft/s",
                       desc="Increment of takeoff rotation speed above engine failure decision speed.")

        self.add_output("Vrot", shape=(1,), units="ft/s",
                        desc="Speed at which takeoff rotation should be initiated.")

        # Constant partials
        self.declare_partials(of="Vrot", wrt=["dV1", "dVR"], val=1.0)
        # Partials of nonlinear terms
        self.declare_partials(
            of="Vrot", wrt=[Dynamic.Mission.MASS, "rho", Aircraft.Wing.AREA, "CL_max"])

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        mass, rho, wing_area, CL_max, dV1, dVR = inputs.values()
        outputs["Vrot"] = ((2 * mass * GRAV_ENGLISH_LBM) /
                           (rho * wing_area * CL_max))**0.5 + dV1 + dVR

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        mass, rho, wing_area, CL_max, dV1, dVR = inputs.values()
        K = 0.5 * ((2 * mass * GRAV_ENGLISH_LBM) /
                   (rho * wing_area * CL_max)) ** 0.5

        partials["Vrot", Dynamic.Mission.MASS] = K / mass
        partials["Vrot", "rho"] = -K / rho
        partials["Vrot", Aircraft.Wing.AREA] = -K / wing_area
        partials["Vrot", "CL_max"] = -K / CL_max



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.fin import FinMass
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import do_validation_test, print_case
from aviary.variable_info.variables import Aircraft, Mission

# TODO - None of the flops cases have any fins.

fin_test_data = {}
fin_test_data['1'] = AviaryValues({
    Aircraft.Fins.NUM_FINS: (1, 'unitless'),
    Mission.Design.GROSS_MASS: (100000, 'lbm'),
    Aircraft.Fins.TAPER_RATIO: (0.3300, 'unitless'),
    Aircraft.Fins.AREA: (250.00, 'ft**2'),
    Aircraft.Fins.MASS_SCALER: (1.0, 'unitless'),
    Aircraft.Fins.MASS: (917.228, 'lbm')
})

fin_data_sets = [key for key in fin_test_data]


class FinMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(fin_data_sets,
                          name_func=print_case)
    def test_case(self, case_name):
        validation_data = fin_test_data[case_name]
        prob = self.prob

        prob.model.add_subsystem(
            "fin",
            FinMass(aviary_options=validation_data),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=validation_data,
                           output_validation_data=validation_data,
                           input_keys=[Mission.Design.GROSS_MASS,
                                       Aircraft.Fins.TAPER_RATIO,
                                       Aircraft.Fins.AREA,
                                       Aircraft.Fins.MASS_SCALER],
                           output_keys=Aircraft.Fins.MASS)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



'''
Define utilities for calculating the effect the ground has on lift-to-drag ratio for
aircraft flying in close proximity to the ground.

References
----------
.. [1] DeYoung, John. "Advanced Supersonic Technology Concept Study Reference
    Characteristics," NASA Contractor Report 132374.
'''
import numpy as np
import openmdao.api as om

from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic


class GroundEffect(om.ExplicitComponent):
    '''
    Define a component for adjusting lift and drag to accomodate ground effect.

    Note, it is an error (no diagnostic) to try to calculate ground effect for an
    aircraft flying "underground" - current `altitude` must ALWAYS be greater than or
    equal to `ground_altitude`.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

        options.declare(
            'ground_altitude', default=0., types=float,
            desc='true altitude of the ground from mean sea level (m)'
        )

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('angle_of_attack', val=np.zeros(nn), units='rad')

        add_aviary_input(self, Dynamic.Mission.ALTITUDE, np.zeros(nn), units='m')

        add_aviary_input(self, Dynamic.Mission.FLIGHT_PATH_ANGLE,
                         val=np.zeros(nn), units='rad')

        self.add_input(
            'minimum_drag_coefficient', 0.0,
            desc='coefficient of minimum drag', units='unitless'
        )

        self.add_input(
            'base_lift_coefficient', val=np.ones(nn),
            desc='coefficient of lift without ground effect', units='unitless'
        )

        self.add_input(
            'base_drag_coefficient', val=np.ones(nn),
            desc='coefficient of drag without ground effect', units='unitless'
        )

        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=1.)
        add_aviary_input(self, Aircraft.Wing.HEIGHT, val=1., units='m')
        add_aviary_input(self, Aircraft.Wing.SPAN, val=1., units='m')

        self.add_output(
            'lift_coefficient', val=np.ones(nn),
            desc='coefficient of lift with ground effect', units='unitless'
        )

        self.add_output(
            'drag_coefficient', val=np.ones(nn),
            desc='coefficient of drag with ground effect', units='unitless'
        )

    def setup_partials(self):
        rows_cols = np.arange(self.options['num_nodes'])

        self.declare_partials(
            'lift_coefficient',
            [Aircraft.Wing.ASPECT_RATIO, Aircraft.Wing.HEIGHT, Aircraft.Wing.SPAN]
        )

        self.declare_partials(
            'lift_coefficient', [Dynamic.Mission.ALTITUDE, 'base_lift_coefficient'],
            rows=rows_cols, cols=rows_cols
        )

        self.declare_partials(
            'lift_coefficient',
            [
                'angle_of_attack', Dynamic.Mission.FLIGHT_PATH_ANGLE, 'minimum_drag_coefficient',
                'base_drag_coefficient',
            ],
            dependent=False
        )

        self.declare_partials(
            'drag_coefficient',
            [Aircraft.Wing.ASPECT_RATIO, Aircraft.Wing.HEIGHT, Aircraft.Wing.SPAN]
        )

        self.declare_partials(
            'drag_coefficient',
            [
                'angle_of_attack', Dynamic.Mission.ALTITUDE, Dynamic.Mission.FLIGHT_PATH_ANGLE,
                'base_drag_coefficient', 'base_lift_coefficient'
            ],
            rows=rows_cols, cols=rows_cols
        )

        self.declare_partials('drag_coefficient', 'minimum_drag_coefficient',
                              rows=rows_cols,
                              cols=np.zeros(self.options['num_nodes'])
                              )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        options = self.options

        ground_altitude = options['ground_altitude']

        angle_of_attack = inputs['angle_of_attack']
        altitude = inputs[Dynamic.Mission.ALTITUDE]
        flight_path_angle = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
        minimum_drag_coefficient = inputs['minimum_drag_coefficient']
        base_lift_coefficient = inputs['base_lift_coefficient']
        base_drag_coefficient = inputs['base_drag_coefficient']
        aspect_ratio = inputs[Aircraft.Wing.ASPECT_RATIO]
        height = inputs[Aircraft.Wing.HEIGHT]
        span = inputs[Aircraft.Wing.SPAN]

        ground_effect_state = ((altitude - ground_altitude) + height) / span
        height_factor = np.ones_like(ground_effect_state)

        idx = np.where(ground_effect_state > 1.0)
        height_factor[idx] = 10. * (1.1 - ground_effect_state[idx])

        aspect_ratio_term = (6. + aspect_ratio)**2 / (36. + aspect_ratio)
        ground_effect_term0 = 32. * (ground_effect_state * aspect_ratio_term)**2 + 1.

        lift_coeff_factor_denom = (
            ground_effect_term0 - 0.5 + 4. * ground_effect_state * aspect_ratio_term
            * np.sqrt(ground_effect_term0)
        )

        lift_coeff_factor = 1. + height_factor / lift_coeff_factor_denom

        lift_coefficient = base_lift_coefficient * lift_coeff_factor

        combined_angle = angle_of_attack + flight_path_angle
        ground_effect_term1 = 1. + 32. * ground_effect_state**2

        drag_coeff_factor_denom = \
            4. * ground_effect_state * np.sqrt(ground_effect_term1) + ground_effect_term1

        drag_coeff_factor = 1. - height_factor / drag_coeff_factor_denom

        drag_coefficient = (
            minimum_drag_coefficient
            + lift_coeff_factor * drag_coeff_factor
            * (base_drag_coefficient - minimum_drag_coefficient)
            + combined_angle * base_lift_coefficient * (lift_coeff_factor - 1.)
        )

        # Check for out of ground effect.
        idx = np.where(ground_effect_state > 1.1)
        drag_coefficient[idx] = base_drag_coefficient[idx]
        lift_coefficient[idx] = base_lift_coefficient[idx]

        outputs['lift_coefficient'] = lift_coefficient
        outputs['drag_coefficient'] = drag_coefficient

    def compute_partials(self, inputs, J, discrete_inputs=None):
        options = self.options

        ground_altitude = options['ground_altitude']

        angle_of_attack = inputs['angle_of_attack']
        altitude = inputs[Dynamic.Mission.ALTITUDE]
        flight_path_angle = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
        minimum_drag_coefficient = inputs['minimum_drag_coefficient']
        base_lift_coefficient = inputs['base_lift_coefficient']
        base_drag_coefficient = inputs['base_drag_coefficient']
        aspect_ratio = inputs[Aircraft.Wing.ASPECT_RATIO]
        height = inputs[Aircraft.Wing.HEIGHT]
        span = inputs[Aircraft.Wing.SPAN]

        # region lift_coefficient wrt [altitude, base_lift_coefficient]
        ground_effect_state = ((altitude - ground_altitude) + height) / span

        d_ges_alt = 1. / span

        height_factor = np.ones_like(ground_effect_state)
        d_hf_alt = np.zeros_like(ground_effect_state)

        idx = np.where(ground_effect_state > 1.0)
        height_factor[idx] = 10. * (1.1 - ground_effect_state[idx])
        d_hf_alt[idx] = -10. / span

        aspect_ratio_term = (6. + aspect_ratio)**2 / (36. + aspect_ratio)
        ground_effect_term0 = 32. * (ground_effect_state * aspect_ratio_term)**2 + 1.

        d_get0_alt = (64. * ground_effect_state * aspect_ratio_term**2) * d_ges_alt

        sqrt_get0 = np.sqrt(ground_effect_term0)

        lift_coeff_factor_denom = (
            ground_effect_term0 - 0.5
            + 4. * ground_effect_state * aspect_ratio_term * sqrt_get0
        )

        d_lcfd_alt = (
            d_get0_alt
            + 4. * aspect_ratio_term * (
                d_ges_alt * sqrt_get0
                + ground_effect_state * (0.5 / sqrt_get0) * d_get0_alt
            )
        )

        lift_coeff_factor = 1. + height_factor / lift_coeff_factor_denom

        d_lcf_alt = (
            (d_hf_alt * lift_coeff_factor_denom) - (height_factor * d_lcfd_alt)
        ) / lift_coeff_factor_denom**2

        J['lift_coefficient', Dynamic.Mission.ALTITUDE] = base_lift_coefficient * d_lcf_alt

        J['lift_coefficient', 'base_lift_coefficient'] = lift_coeff_factor
        # endregion lift_coefficient wrt [altitude, base_lift_coefficient]

        # region lift_coefficient wrt Aircraft.Wing.ASPECT_RATIO
        f = (6. + aspect_ratio)**2
        d_f = 2. * (6. + aspect_ratio)
        g = (36. + aspect_ratio)
        d_g = 1.

        d_art_ar = ((d_f * g) - (f * d_g)) / g**2

        d_get0_ar = 64. * (ground_effect_state**2 * aspect_ratio_term) * d_art_ar

        d_lcfd_ar = d_get0_ar + 4. * ground_effect_state * (
            d_art_ar * sqrt_get0
            + aspect_ratio_term * 0.5 / sqrt_get0 * d_get0_ar
        )

        f = height_factor
        d_f = 0.
        g = lift_coeff_factor_denom
        d_g = d_lcfd_ar

        d_lcf_ar = -f * d_g / g**2

        J['lift_coefficient', Aircraft.Wing.ASPECT_RATIO] = \
            base_lift_coefficient * d_lcf_ar
        # endregion lift_coefficient wrt Aircraft.Wing.ASPECT_RATIO

        J['lift_coefficient', Aircraft.Wing.HEIGHT] = base_lift_coefficient * d_lcf_alt

        # region lift_coefficient wrt Aircraft.Wing.SPAN
        d_ges_b = -((altitude - ground_altitude) + height) / span**2

        d_hf_b = np.zeros_like(ground_effect_state)
        d_hf_b[idx] = -10. * d_ges_b[idx]

        d_get0_b = 64. * (ground_effect_state * aspect_ratio_term**2) * d_ges_b

        d_lcfd_b = (
            d_get0_b + 4. * aspect_ratio_term * (
                (d_ges_b * sqrt_get0)
                + (ground_effect_state * 0.5 / sqrt_get0 * d_get0_b)
            )
        )

        f = height_factor
        d_f = d_hf_b
        g = lift_coeff_factor_denom
        d_g = d_lcfd_b

        d_lcf_b = ((d_f * g) - (f * d_g)) / g**2

        J['lift_coefficient', Aircraft.Wing.SPAN] = base_lift_coefficient * d_lcf_b
        # endregion lift_coefficient wrt Aircraft.Wing.SPAN

        # region drag_coefficient wrt angle_of_attack
        combined_angle = angle_of_attack + flight_path_angle

        d_ca_aoa = 1.

        ground_effect_term1 = 1. + 32. * ground_effect_state**2

        sqrt_get1 = np.sqrt(ground_effect_term1)

        drag_coeff_factor_denom = \
            4. * ground_effect_state * sqrt_get1 + ground_effect_term1

        drag_coeff_factor = 1. - height_factor / drag_coeff_factor_denom

        d_dc_aoa = base_lift_coefficient * (lift_coeff_factor - 1.) * d_ca_aoa

        J['drag_coefficient', 'angle_of_attack'] = d_dc_aoa
        # endregion drag_coefficient wrt angle_of_attack

        # region drag_coefficient wrt flight_path_angle
        d_ca_fpa = 1.

        d_dc_fpa = base_lift_coefficient * (lift_coeff_factor - 1.) * d_ca_fpa

        J['drag_coefficient', Dynamic.Mission.FLIGHT_PATH_ANGLE] = d_dc_fpa
        # endregion drag_coefficient wrt flight_path_angle

        # region drag_coefficient wrt altitude
        d_get1_alt = 64. * ground_effect_state * d_ges_alt

        f = ground_effect_state
        d_f = d_ges_alt
        g = sqrt_get1
        d_g = 0.5 * d_get1_alt / sqrt_get1

        d_dcfd_alt = 4. * ((d_f * g) + (f * d_g)) + d_get1_alt

        f = height_factor
        d_f = d_hf_alt
        g = drag_coeff_factor_denom
        d_g = d_dcfd_alt

        d_dcf_alt = -((d_f * g) - (f * d_g)) / g**2

        f = lift_coeff_factor
        d_f = d_lcf_alt
        g = drag_coeff_factor
        d_g = d_dcf_alt

        d_dc_alt = (
            ((d_f * g) + (d_g * f)) * (base_drag_coefficient - minimum_drag_coefficient)
            + combined_angle * base_lift_coefficient * d_lcf_alt
        )

        J['drag_coefficient', Dynamic.Mission.ALTITUDE] = d_dc_alt
        # endregion drag_coefficient wrt altitude

        # region drag_coefficient wrt minimum_drag_coefficient
        d_dc_mdc = 1. - lift_coeff_factor * drag_coeff_factor

        J['drag_coefficient', 'minimum_drag_coefficient'] = d_dc_mdc
        # endregion drag_coefficient wrt minimum_drag_coefficient

        # region drag_coefficient wrt base_lift_coefficient
        d_dc_blc = combined_angle * (lift_coeff_factor - 1.)

        J['drag_coefficient', 'base_lift_coefficient'] = d_dc_blc
        # endregion drag_coefficient wrt base_lift_coefficient

        # region drag_coefficient wrt base_drag_coefficient
        d_dc_bdc = lift_coeff_factor * drag_coeff_factor

        J['drag_coefficient', 'base_drag_coefficient'] = d_dc_bdc
        # endregion drag_coefficient wrt base_drag_coefficient

        # region drag_coefficient wrt Aircraft.Wing.ASPECT_RATIO
        d_dc_ar = (
            drag_coeff_factor * (base_drag_coefficient - minimum_drag_coefficient)
            + combined_angle * base_lift_coefficient
        ) * d_lcf_ar

        J['drag_coefficient', Aircraft.Wing.ASPECT_RATIO] = d_dc_ar
        # endregion drag_coefficient wrt Aircraft.Wing.ASPECT_RATIO

        J['drag_coefficient', Aircraft.Wing.HEIGHT] = d_dc_alt

        # region drag_coefficient wrt Aircraft.Wing.SPAN
        d_get1_b = 64. * ground_effect_state * d_ges_b

        f = ground_effect_state
        d_f = d_ges_b
        g = sqrt_get1
        d_g = 0.5 * d_get1_b / sqrt_get1

        d_dcfd_b = 4. * ((d_f * g) + (f * d_g)) + d_get1_b

        f = height_factor
        d_f = d_hf_b
        g = drag_coeff_factor_denom
        d_g = d_dcfd_b

        d_dcf_b = -((d_f * g) - (f * d_g)) / g**2

        f = lift_coeff_factor
        d_f = d_lcf_b
        g = drag_coeff_factor
        d_g = d_dcf_b

        d_dc_b = (
            ((d_f * g) + (f * d_g)) * (base_drag_coefficient - minimum_drag_coefficient)
            + combined_angle * base_lift_coefficient * d_lcf_b
        )

        J['drag_coefficient', Aircraft.Wing.SPAN] = d_dc_b
        # endregion drag_coefficient wrt Aircraft.Wing.SPAN

        # Check for out of ground effect.
        idx = np.where(ground_effect_state > 1.1)
        if idx:
            J['drag_coefficient', Dynamic.Mission.ALTITUDE][idx] = 0.0
            J['drag_coefficient', 'minimum_drag_coefficient'][idx] = 0.0
            J['drag_coefficient', 'base_lift_coefficient'][idx] = 0.0
            J['drag_coefficient', 'base_drag_coefficient'][idx] = 1.0
            J['drag_coefficient', Aircraft.Wing.ASPECT_RATIO][idx] = 0.0
            J['drag_coefficient', Aircraft.Wing.HEIGHT][idx] = 0.0
            J['drag_coefficient', Aircraft.Wing.SPAN][idx] = 0.0
            J['drag_coefficient', 'angle_of_attack'][idx] = 0.0
            J['drag_coefficient', Dynamic.Mission.FLIGHT_PATH_ANGLE][idx] = 0.0

            J['lift_coefficient', Dynamic.Mission.ALTITUDE][idx] = 0.0
            J['lift_coefficient', 'base_lift_coefficient'][idx] = 1.0
            J['lift_coefficient', Aircraft.Wing.ASPECT_RATIO][idx] = 0.0
            J['lift_coefficient', Aircraft.Wing.HEIGHT][idx] = 0.0
            J['lift_coefficient', Aircraft.Wing.SPAN][idx] = 0.0
